#include "test.h"

#define OK(expr) \
{ \
  (expr); \
  printf("%s ok %d\n", #expr, __LINE__); \
  fflush(stdout); \
}

#define WHITE 255, 255, 255
#define RED   255,   0,   0
#define GREEN   0, 255,   0
#define BLUE    0,   0, 255
#define BLACK   0,   0,   0
#define CUNK	BLACK

#define ORTHO_GL 1.f

#define LISTCOUNT_GL 5
#define SKELLIST_GL gl_listbase + 0
#define REALLIST_GL gl_listbase + 1
#define ORTSLIST_GL gl_listbase + 2
#define TEXTLIST_GL gl_listbase + 3
#define HELPLIST_GL gl_listbase + 4

#define ANGLE_INCREMENT 10.f

static GLsizei width  = 640;
static GLsizei height = 480;

static GLuint gl_texttex;
static GLuint gl_listbase;
static GLuint * gl_texnames;
static GLint filter = GL_NEAREST;

static struct SntrTag * sntr;
static struct MotnTag * motn;
static struct SktrTag * sktr;
static struct RetrTag * retr;
static struct MotnTag * cur_motn;
static struct MotnTag * root_motn;

static int verbose;
static int scale;
static int idle_play   = 0;
static int render_skel = 0;
static int render_orts = 0;
static int render_real = 1;
static int render_text = 0;
static int render_help = 1;
static uint * ress;

static void orts_compile(GLuint list, struct RetrTag * retr)
{
  #define x 1.f / retr->rs * 0.5f
  #define y 1.f / retr->rs * 0.5f
  #define z 1.f / retr->rs * 0.5f
  glNewList(list, GL_COMPILE);
    glDisable(GL_TEXTURE_2D);
    glColor4ub(BLUE, 255);

    glBegin(GL_LINES);
      glVertex3f(0.f, 0.f, 0.f);
      glVertex3f(  x, 0.f, 0.f);
      glVertex3f(0.f, 0.f, 0.f);
      glVertex3f(0.f,   y, 0.f);
      glVertex3f(0.f, 0.f, 0.f);
      glVertex3f(0.f, 0.f,   z);
    glEnd();

  glEndList();
  #undef x
  #undef y
  #undef z
}

static void skel_compile(GLuint list, struct SktrTag * sktr)
{
  int i;

  glNewList(list, GL_COMPILE);
    glColor4ub(RED, 255);

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);

    glBegin(GL_LINES);
      for (i = 0; i < sktr->skel->LimbCount; i++)
      {
        glVertex3fv(&sktr->LimbListS[i * 3]);
        glVertex3fv(&sktr->LimbListE[i * 3]);
      }
    glEnd();

    glColor4ub(GREEN, 255);
    glPointSize(2.0);
    glBegin(GL_POINTS);
      for (i = 0; i < sktr->skel->LimbCount; i++)
      {
        glVertex3fv(&sktr->LimbListE[i * 3]);
      }
    glEnd();
    glPointSize(1.0);

    glEnable(GL_DEPTH_TEST);
  glEndList();
}

/*
 * ripoff from some GLUT demo
 */
static char * gluthelp = "Since (glut)idle differs mucho from Win32(aux) counterpart\n"
"And since i have no intention to mess with glut timers\n"
"I suggest you to play motions pressing 'f'(somewhat faster)\n";

static void text_compile(GLuint list, int x, int y, const char * string)
{                                                                                                   
    int lines;
    char * p;
    char * s = (char *) string;

    glNewList(list, GL_COMPILE);
        
	glDisable(GL_DEPTH_TEST);                                                                       
	glMatrixMode(GL_PROJECTION);                                                                    
	glPushMatrix();                                                                                 
	glLoadIdentity();                                                                               
	glOrtho(0, glutGet(GLUT_WINDOW_WIDTH),                                                          
		0, glutGet(GLUT_WINDOW_HEIGHT), -1, 1);                                                 
	glMatrixMode(GL_MODELVIEW);                                                                     
	glPushMatrix();                                                                                 
	glLoadIdentity();                                                                               
	glColor3ub(0, 0, 0);                                                                            
	glRasterPos2i(x+1, y-1);                                                                        
	for(p = s, lines = 0; *p; p++) 
	{
	    if (*p == '\n') 
	    {
		lines++;                                                                                
		glRasterPos2i(x+1, y-1-(lines*18));                                                     
	    }                                                                                           
	    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *p);                                          
	}                                                                                               
	glColor3ub(128, 0, 255);                                                                        
	glRasterPos2i(x, y);                                                                            
	for(p = s, lines = 0; *p; p++) 
	{
	    if (*p == '\n') 
	    {
		lines++;                                                                                
		glRasterPos2i(x, y-(lines*18));                                                         
	    }                                                                                           
	    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *p);                                          
	}        
	glMatrixMode(GL_PROJECTION);                                                                    
	glPopMatrix();                                                                                  
	glMatrixMode(GL_MODELVIEW);                                                                     
	glPopMatrix();                                                                                  
	glEnable(GL_DEPTH_TEST);                                                                        
    glEndList();
}

/*
 * this routine is a horrible mess, oh well
 */
static void real_compile(GLuint list, struct RetrTag * retr)
{
  int i;
  GLuint curname = -1;
  struct RealTag * real = retr->real;

  #if 0
  for (i = 0; i < real->VertexCount; i++)
  {
    retr->VertexList[i * 3 + 0] = (float) real->VertexList[i].x;
    retr->VertexList[i * 3 + 1] = (float) real->VertexList[i].y;
    retr->VertexList[i * 3 + 2] = (float) real->VertexList[i].z;
  }
  #endif

  glNewList(list, GL_COMPILE);

    glDisable(GL_TEXTURE_2D);
    glColor4ub(WHITE, 255);

    /*
     * in BioForge lines are black, but i figured that
     * it would be somewhat hard to notice black lines
     * on black field...
     */

    if (real->LineFaceCount > 0)
    {
      glBegin(GL_LINES);
      for (i = 0; i < real->LineFaceCount; i++)
      {
        float * p1 = &retr->VertexList[real->LineFaceList[i].a * 3];
        float * p2 = &retr->VertexList[real->LineFaceList[i].b * 3];
        glVertex3fv(p1);
        glVertex3fv(p2);
      }
      glEnd();
    }

    glEnable(GL_TEXTURE_2D);

    for (i = 0; i < real->TrisFaceCount; i++)
    { 
      GLuint texname;
      float * p1, * p2, * p3;
      GLfloat w, h;
      struct MapTag * map = &retr->MapList[i];      

      p1 = &retr->VertexList[real->TrisFaceList[i].a * 3];
      p2 = &retr->VertexList[real->TrisFaceList[i].b * 3];
      p3 = &retr->VertexList[real->TrisFaceList[i].c * 3];

      if (map->MapType == MAP_TYPE_NONE) 
      {
        /*
         * i have no fucking idea what polys without
         * attached mapping mean
         */

        glDisable(GL_TEXTURE_2D);
        glColor3ub(CUNK);
        glBegin(GL_TRIANGLES);
          glVertex3fv(p1);
          glVertex3fv(p2);
          glVertex3fv(p3);
        glEnd();
        glEnable(GL_TEXTURE_2D);
        continue;
      }
      
      texname = gl_texnames[map->Map.TrisMap->textureid];
      if (curname != texname)
      {
        glBindTexture(GL_TEXTURE_2D, texname);
        curname = texname;
        w = (GLfloat) ress[map->Map.TrisMap->textureid * 2 + 0];
        h = (GLfloat) ress[map->Map.TrisMap->textureid * 2 + 1];
      }

      #define W(p) (GLfloat) ((GLfloat) map->Map.TrisMap->p / w)
      #define H(p) (GLfloat) ((GLfloat) map->Map.TrisMap->p / h)

      #if 0
      glDisable(GL_TEXTURE_2D);
      glColor3d(RED);
      #endif

      glBegin(GL_TRIANGLES);
        glTexCoord2f(W(ax), H(ay));
        glVertex3fv(p1);
  
        glTexCoord2f(W(bx), H(by));
        glVertex3fv(p2);

        glTexCoord2f(W(cx), H(cy));
        glVertex3fv(p3);
      glEnd();

      #undef W
      #undef H
    }

    glEnable(GL_TEXTURE_2D);

    for (i = 0; i < real->QuadFaceCount; i++)
    { 
      GLuint texname;
      float * p1, * p2, * p3, * p4;
      GLfloat w, h;
      struct MapTag * map = &retr->MapList[i + real->TrisFaceCount];      

      p1 = &retr->VertexList[real->QuadFaceList[i].a * 3];
      p2 = &retr->VertexList[real->QuadFaceList[i].b * 3];
      p3 = &retr->VertexList[real->QuadFaceList[i].c * 3];
      p4 = &retr->VertexList[real->QuadFaceList[i].d * 3];

      if (map->MapType == MAP_TYPE_NONE) 
      {
        /*
         * i have no fucking idea what polys without
         * attached mapping mean
         */

        glDisable(GL_TEXTURE_2D);
        glColor3ub(CUNK);
        glBegin(GL_QUADS);
          glVertex3fv(p1);
          glVertex3fv(p2);
          glVertex3fv(p3);
          glVertex3fv(p4);
        glEnd();
        glEnable(GL_TEXTURE_2D);
        continue;
      }
      
      texname = gl_texnames[map->Map.QuadMap->textureid];
      if (curname != texname)
      {
        glBindTexture(GL_TEXTURE_2D, texname);
        curname = texname;
        w = (GLfloat) ress[map->Map.QuadMap->textureid * 2 + 0];
        h = (GLfloat) ress[map->Map.QuadMap->textureid * 2 + 1];
      }

      #define W(p) (GLfloat) ((GLfloat) map->Map.QuadMap->p / w)
      #define H(p) (GLfloat) ((GLfloat) map->Map.QuadMap->p / h)

      #if 0
      glDisable(GL_TEXTURE_2D);
      glColor3ub(RED);
      #endif

      glBegin(GL_QUADS);
        glTexCoord2f(W(ax), H(ay));
        glVertex3fv(p1);
  
        glTexCoord2f(W(bx), H(by));
        glVertex3fv(p2);

        glTexCoord2f(W(cx), H(cy));
        glVertex3fv(p3);

        glTexCoord2f(W(dx), H(dy));
        glVertex3fv(p4);
      glEnd();

      #undef W
      #undef H
    }

  glEndList();
}

static uint getp2(uint v)
{
  uint i = 0;
  uint j = 0;
  while (i < v)
  {
    i = 2 << j;
    j++;
  }
  return i;
}

static void real_load_textures(struct RetrTag * retr)
{
  int i;
  uint * res;
  struct RealTag * real;

  real = retr->real;
  gl_texnames = malloc(sizeof(GLuint) * real->TextureCount);
  ress = malloc(sizeof(uint) * real->TextureCount * 2);
  res = ress;

  glGenTextures(real->TextureCount, gl_texnames);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  for (i = 0; i < real->TextureCount; i++)
  {
    uchar * image8 = real->TextureList[i].bitmap;
    uint xres = real->TextureList[i].xres;
    uint yres = real->TextureList[i].yres;
    uint xres2 = getp2(xres);
    uint yres2 = getp2(yres);
    uchar * imageRGBA = malloc(xres2 * yres2 * 4);
    uint x, y;

    memset(imageRGBA, 0x80, xres2 * yres2 * 4);
    for (y = 0; y < yres; y++)
    {
      uint j = y * xres2 * 4;
      for (x = 0; x < xres; x++)
      {
        uint index = (uint) *image8++;
        if (index == 0xff)
        {
          imageRGBA[j++] = 0;
          imageRGBA[j++] = 0;
          imageRGBA[j++] = 0;
          imageRGBA[j++] = 0;
        } else
        {
          #define REAL_PALR(I) retr->palette[I * 3 + 0]
          #define REAL_PALG(I) retr->palette[I * 3 + 1]
          #define REAL_PALB(I) retr->palette[I * 3 + 2]
          #define REAL_PALA(I) 255

          imageRGBA[j++] = REAL_PALR(index);
          imageRGBA[j++] = REAL_PALG(index);
          imageRGBA[j++] = REAL_PALB(index);
          imageRGBA[j++] = REAL_PALA(index);

          #undef REAL_PALR
          #undef REAL_PALG
          #undef REAL_PALB
          #undef REAL_PALA
        }
      }
    }

    if (scale)
    {
      uchar * scaledImage = malloc(xres2 * yres2 * 4);

      glPixelStorei(GL_UNPACK_ROW_LENGTH, xres2);
      gluScaleImage(GL_RGBA, xres,  yres,  GL_UNSIGNED_BYTE, imageRGBA,
                             xres2, yres2, GL_UNSIGNED_BYTE, scaledImage);
      *res++ = xres;
      *res++ = yres;

      free(imageRGBA);
      imageRGBA = scaledImage;
      glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    }
    else
    {
      *res++ = xres2;
      *res++ = yres2;
    }

    glBindTexture(GL_TEXTURE_2D, gl_texnames[i]);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, xres2, yres2, 0, GL_RGBA,
                 GL_UNSIGNED_BYTE, imageRGBA);

    free(imageRGBA);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  }
}

static void init_gl(struct RetrTag * retr)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glRotatef(90.f, 1.f, 0.f, 0.f);
  glScalef(retr->rs, retr->rs, retr->rs);

  glOrtho(-ORTHO_GL, ORTHO_GL, -ORTHO_GL, ORTHO_GL, -ORTHO_GL, ORTHO_GL); 

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(-retr->xc, -retr->yc, -retr->zc);
  
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_FLAT);


  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);


  glEnable(GL_DEPTH_TEST);
  glClearDepth(1.0);
  glDepthFunc(GL_LESS);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glEnable(GL_POINT_SMOOTH);

  gl_listbase = glGenLists(LISTCOUNT_GL);
}

static void CALLBACK glut_reshape(GLsizei w, GLsizei h)
{
  width = w;
  height = h; 
  text_compile(TEXTLIST_GL, 0, height - 20, gluthelp);
  text_compile(HELPLIST_GL, 0, height - 20, help);
  glViewport(0, 0, w, h);
}

static void CALLBACK glut_display(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  if (render_real) glCallList(REALLIST_GL);
  if (render_skel) glCallList(SKELLIST_GL);
  if (render_orts) glCallList(ORTSLIST_GL);
  if (render_text) glCallList(TEXTLIST_GL);
  if (render_help) glCallList(HELPLIST_GL);
  glutSwapBuffers();
}

static void CALLBACK glutkey_left(void)
{
  glMatrixMode(GL_PROJECTION);
  glRotatef(-ANGLE_INCREMENT, 0.f, 1.f, 0.f);
}

static void CALLBACK glutkey_right(void)
{
  glMatrixMode(GL_PROJECTION);
  glRotatef(ANGLE_INCREMENT, 0.f, 1.f, 0.f);
}

static void CALLBACK glutkey_up(void)
{
  glMatrixMode(GL_PROJECTION);
  glRotatef(-ANGLE_INCREMENT, 1.f, 0.f, 0.f);
}

static void CALLBACK glutkey_down(void)
{
  glMatrixMode(GL_PROJECTION);
  glRotatef(ANGLE_INCREMENT, 1.f, 0.f, 0.f);
}

static void CALLBACK glutkey_z(void)
{
  glMatrixMode(GL_PROJECTION);
  glRotatef(-ANGLE_INCREMENT, 0.f, 0.f, 1.f);
}

static void CALLBACK glutkey_x(void)
{
  glMatrixMode(GL_PROJECTION);
  glRotatef(ANGLE_INCREMENT, 0.f, 0.f, 1.f);
}

static void CALLBACK glutkey_m(void)
{
  render_real = !render_real;
}

static int poseno = 0;
static int first = 1;

static void CALLBACK glut_idle(void)
{
  if (idle_play)
  {
    if (first)
    {
      motn_start(cur_motn, sktr, poseno);
      first = 0;
    }
    if (motn_next(sktr))
    {
      if (verbose) printf("motion %s pose %d\n", cur_motn->motn.name, poseno);
      poseno++;
      if (poseno > cur_motn->motn.PoseCount - 1) poseno = 0;
      motn_start(cur_motn, sktr, poseno);
    } 
    skel_next(sktr, retr, sntr);
    real_compile(REALLIST_GL, retr);
    skel_compile(SKELLIST_GL, sktr);
    glutPostRedisplay();
  }
}

static void CALLBACK glutkey_f(void)
{
  if (first)
  {
    motn_start(cur_motn, sktr, poseno);
    first = 0;
  }
  if (motn_next(sktr))
  {
    if (verbose) printf("motion %s pose %d\n", cur_motn->motn.name, poseno);
    poseno++;
    if (poseno > cur_motn->motn.PoseCount - 1) poseno = 0;
    motn_start(cur_motn, sktr, poseno);
  }
  skel_next(sktr, retr, sntr);
  real_compile(REALLIST_GL, retr);
  skel_compile(SKELLIST_GL, sktr);
}

static void CALLBACK glutkey_r(void)
{
}

static void CALLBACK glutkey_n(void)
{
  cur_motn = cur_motn->link;
  if (!cur_motn) cur_motn = root_motn->link;
  poseno = 0;
}

static void CALLBACK glutkey_s(void)
{
  render_skel = !render_skel;
}

static void CALLBACK glutkey_o(void)
{
  render_orts = !render_orts;
}

static void CALLBACK glutkey_w(void)
{
  int i;
  FILE * fp;
  uchar tga[18];
  GLubyte * pixels;
  GLenum error;
  int size = width * height * 3;
  pixels = malloc(size);

  glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, pixels);
  error = glGetError();
  if (error != GL_NO_ERROR)
    fprintf(stderr, "glReadPixels %s\n", gluErrorString(error));

  memset(tga, 0, 18);
  tga[2]  = 2;
  tga[12] = (uchar) (width & 0xff);
  tga[13] = (uchar) (width >> 8);
  tga[14] = (uchar) (height & 0xff);
  tga[15] = (uchar) (height >> 8);
  tga[16] = 24;
  
  for (i = 0; i < size / 3; i++)
  {
    GLubyte r = pixels[i * 3 + 0];
    GLubyte g = pixels[i * 3 + 1];
    GLubyte b = pixels[i * 3 + 2];
    pixels[i * 3 + 0] = b;
    pixels[i * 3 + 1] = g;
    pixels[i * 3 + 2] = r;
  }

  fp = fopen("sshot.tga", "wb");
  fwrite(tga, 18, 1, fp);
  fwrite(pixels, size, 1, fp);
  fclose(fp);

  free(pixels);
}

static void CALLBACK glutkey_a(void)
{
  if (glIsEnabled(GL_ALPHA_TEST))
  {
    glDisable(GL_ALPHA_TEST);
  } else
  {
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_NOTEQUAL, 0.0);
  }
}

static void CALLBACK glutkey_b(void)
{
  int i;

  filter = filter == GL_NEAREST ? GL_LINEAR : GL_NEAREST;

  for (i = 0; i < retr->real->TextureCount; i++)
  {
    glBindTexture(GL_TEXTURE_2D, gl_texnames[i]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
  }
}

static void CALLBACK glutkey_plus(void)
{
  glMatrixMode(GL_PROJECTION);
  glScalef(1.1f, 1.1f, 1.1f);
}

static void CALLBACK glutkey_minus(void)
{
  glMatrixMode(GL_PROJECTION);
  glScalef(1.f / 1.1f, 1.f / 1.1f, 1.f / 1.1f);
}

static void CALLBACK glutkey_p(void)
{
  idle_play = !idle_play;
}

static void CALLBACK glutkey_h(void)
{
  render_help = !render_help;
}

static void CALLBACK glutkey_v(void)
{
#ifdef _WIN32
  MessageBoxA(auxGetHWND(), salsa_version(), "Version", MB_OK | MB_ICONEXCLAMATION | MB_APPLMODAL);
#else
  printf("%s\n", salsa_version());
#endif
}

static void show_help(void)
{
#ifdef _WIN32
  MessageBoxA(NULL, help, "Help", MB_OK | MB_ICONINFORMATION | MB_APPLMODAL);
#else
  printf("%s\n", help);
#endif
}

static void info()
{
  int i;
  struct MotnTag * motn;

  printf("Shape %s:\n", retr->real->name);
  printf(" VertexCount  : %d\n", retr->real->VertexCount);
  printf(" LineFaceCount: %d\n", retr->real->LineFaceCount);
  printf(" TrisFaceCount: %d\n", retr->real->TrisFaceCount);
  printf(" QuadFaceCount: %d\n", retr->real->QuadFaceCount);
  printf(" TrisMapCount : %d\n", retr->real->TrisMapCount);
  printf(" QuadMapCount : %d\n", retr->real->QuadMapCount);
  printf(" TextureCount : %d\n", retr->real->TextureCount); 

  printf("Skeleton:\n");
  printf(" LimbCount    : %d\n", sktr->skel->LimbCount);

  for (i = 0, motn = motn_root(); motn; motn = motn->link)  i++;
  printf("Motions:\n");
  printf(" MotionCount  : %d\n", i);
}

static void info_salsa()
{
  printf("SALSA Information\n");
  printf("%s\n", salsa_version());
}

static void info_gl()
{
#if 0
  char buf[256];
  sprintf(buf,
          "Vendor   : %s\n"
          "Renderer : %s\n"
          "Version  : %s\n",
          (char *) glGetString(GL_VENDOR),
          (char *) glGetString(GL_RENDERER),
          (char *) glGetString(GL_VERSION));
  MessageBoxA(NULL, buf, "OpenGL", MB_OK | MB_ICONINFORMATION | MB_APPLMODAL);
#else
  printf("OpenGL Information\n");
  printf("Vendor   : %s\n", (char *) glGetString(GL_VENDOR));
  printf("Renderer : %s\n", (char *) glGetString(GL_RENDERER));
  printf("Version  : %s\n", (char *) glGetString(GL_VERSION));
#endif
}

static char * open_bed(char * name)
{
#ifdef _WIN32
  static char file[_MAX_PATH + 1];
  static OPENFILENAME ofn = 
  {
    sizeof(ofn), 0, 0,
    "SALSA Actor Definition (*.bed)\0*.bed\0\0\0",
    0, 0, 0,
    file,
    _MAX_PATH,
    0, 0,
    "data",
    NULL,
    OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_EXPLORER,
    0, 0, "bed", 0, 0, 0
  };
  
  struct _stat buf;

  if (_stat(name, &buf) == -1)
    if (GetOpenFileNameA(&ofn)) return file;
  return name
#endif
  return name;
}

static void glut_specialfunc(int key, int x, int y)
{
  switch (key)
  {
  case GLUT_KEY_LEFT:  glutkey_left(); break;
  case GLUT_KEY_RIGHT: glutkey_right(); break;
  case GLUT_KEY_UP:    glutkey_up(); break;
  case GLUT_KEY_DOWN:  glutkey_down(); break;
  default: return;
  }
  glutPostRedisplay();
}

static void glut_keyboardfunc(unsigned char key, int x, int y)
{
    switch (key)
    {
    case 'x': glutkey_x(); break;
    case 'z': glutkey_z(); break;
    case 'm': glutkey_m(); break;
    case 's': glutkey_s(); break;
    case 'o': glutkey_o(); break;
    case 'f': glutkey_f(); break;
    case 'r': glutkey_r(); break;
    case 'w': glutkey_w(); break;
    case 'n': glutkey_n(); break;
    case 'a': glutkey_a(); break;
    case 'b': glutkey_b(); break;
    case 'p': glutkey_p(); break;
    case 'v': glutkey_v(); break;
    case 'h': glutkey_h(); break;
    case '0': glutkey_plus(); break;
    case '9': glutkey_minus(); break;

    /* 
     * well fullscreen is nice no question about it
     * but... how do i restore normal operation?
     *
     case 'F': glutFullScreen(); break;
     * and what's the size of the screen?
     */

    case 'q': 	
    case 0x1b: 
	exit(0); /* do it nice and intellegent ;) */ 

    default: return;
    }
    glutPostRedisplay();
}
	

int main(int argc, char ** argv)
{
  struct BedTag * bed;
  char * bed_file;
  char * lex_bed = "lex.bed";
  char * name_motn = "STAND_2";
  char platform[256];

  args_init(argc, argv);

  verbose = arg_verbose;
  scale = arg_scale_image;

  bed_file = open_bed(!arg_bed_file ? lex_bed : arg_bed_file);

  if (verbose)
    printf("using bed-file %s\n", bed_file);

  bed = bed_init(bed_file);
  sntr = skin_init(bed->skin);
  sktr = skel_init(bed->skel);
  motn = motn_init(bed->motn);
  retr = real_init(bed->real);

  sktr->SlerpStepCount = arg_slerp_steps;

  skin_process(sntr, retr);
  skel_process(sktr, retr, sntr);
  skel_apply(sktr, retr, sntr);

  root_motn = motn_root();

  if (!strcmp(bed_file, lex_bed))
    name_motn = "STAND_2";
  else
    name_motn = arg_mtn_name;

  if (!strcmp(name_motn, "root"))
    cur_motn = root_motn->link;
  else
    cur_motn = motn_find(name_motn);

  if (verbose)
    printf("current motion %s:%d\n",
           cur_motn->motn.name,
           cur_motn->motn.PoseCount);
  
  if (arg_print_names)
  {
    struct MotnTag * motn;
    for (motn = root_motn->link; motn; motn = motn->link)
    {
      printf("%s\n", motn->motn.name);
    }
    fflush(stdout);
  }

  sprintf(platform,
          "SALSA Player " PLATFORM "/" COMPILER " %d.%d, Built "
          __DATE__ " " __TIME__,
          CVERSION);
           
  show_help();
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutCreateWindow(platform);

  if (verbose)
  {
    info_gl();
    info_salsa();
  }

  init_gl(retr);

  real_load_textures(retr);

  orts_compile(ORTSLIST_GL, retr);
  skel_compile(SKELLIST_GL, sktr);
  real_compile(REALLIST_GL, retr);
  
  glutKeyboardFunc(&glut_keyboardfunc);
  glutSpecialFunc(&glut_specialfunc);
  glutReshapeFunc(&glut_reshape);
  glutIdleFunc(&glut_idle);
  glutDisplayFunc(&glut_display);
  glutReshapeWindow(width, height);
  glutMainLoop();

  return 0;
}
