#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bed.h"

static char * bedpath;
static int path_len;

#define MAX_LINE_LEN 256

static void get_path(const char * s)
{
  int i;
  int len;

  bedpath = 0;
  path_len = 0;

  len = strlen(s);
  for (i = len; i > 0; i--)
  {
    if (s[i] == '\\' || s[i] == '/')
    {
      int j;

      path_len = i + 2;

      bedpath = malloc(path_len);
      _CHECK(!bedpath, "malloc(%u) failed", path_len, 0);

      for (j = 0; j <= i; j++)
      {
        bedpath[j] = s[j];
      }
      path_len--;
      bedpath[path_len] = 0;
      break;
    }
  }
}

static int init_name(const char * line, const char * fmt, char ** name)
{
  char temp[MAX_LINE_LEN];

  if (sscanf(line, fmt, temp) == 1)
  {
    int len;

    len = strlen(temp);
    *name = malloc(len + path_len + 1);
    **name = 0;
    if (bedpath) strcpy(*name, bedpath);
    strcat(*name, temp);
    return 1;
  }
  return 0;
}

struct BedTag * bed_init(const char * filename)
{
  struct BedTag * bed;

  FILE * fp;

  fp = fopen(filename, "r");
  _CHECK(!fp, "can't open %s", filename, 0);

  bed = malloc(BEDTAGSZ);
  _CHECK(!bed, "malloc(%u) failed", BEDTAGSZ, 0);
 
  get_path(filename);

  bed->skin = 0;
  bed->skel = 0;
  bed->motn = 0;
  bed->real = 0;

  while (!feof(fp))
  {
    char * c;
    char line[MAX_LINE_LEN];
    int len;

    fgets(line, MAX_LINE_LEN, fp);
    len = strlen(line);
    if (len == 0) continue;

    c = strchr(line, ';');
    if (c) *c = 0;

    len = strlen(line);
    if (len < 6) continue;

    if (init_name(line, "REAL = %s", &bed->real)) continue;
    if (init_name(line, "SKIN = %s", &bed->skin)) continue;
    if (init_name(line, "SKEL = %s", &bed->skel)) continue;
    if (init_name(line, "MOTN = %s", &bed->motn)) continue;
  }

  _CHECK(!bed->skin, "no SKIN found in %s", filename, 0);
  _CHECK(!bed->real, "no REAL found in %s", filename, 0);
  _CHECK(!bed->skel, "no SKEL found in %s", filename, 0);
  _CHECK(!bed->motn, "no MOTN found in %s", filename, 0);
  if (bedpath) free(bedpath);

  fclose(fp);
  return bed;
}

void bed_done(struct BedTag * bed)
{
  if (bed->skin) free(bed->skin);
  if (bed->skel) free(bed->skel);
  if (bed->motn) free(bed->motn);
  if (bed->real) free(bed->real);
  free(bed);
}

#ifdef _TEST
int main(void)
{
  bed_init("lex.bed");
}
#endif
