#include "config.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "motn.h"
#include "skel.h"
#include "vec/qtr.h"

static struct MotnTag * root = 0;

struct MotnTag * motn_init(const char * name)
{
  root = mtn_init(stream_init(name));
  return root;
}

void motn_done(struct MotnTag * motn)
{
  /* 
   * better pass a very root motn here otherwise some link will point to nada
   */

  while (motn)
  {
    struct MotnTag * temp;

    if (motn->DeltaList) free(motn->DeltaList);
    if (motn->AngleList) free(motn->AngleList);
    temp = motn;
    motn = motn->link;
    if (temp) free(temp);
  }
}

struct MotnTag * motn_find(const char * name)
{
  struct MotnTag * motn;
  for (motn = root; motn; motn = motn->link)
  {
    if (!strcmp(motn->motn.name, name)) return motn;
  }

  _CHECK(!motn, "motion %s not found", name, 0);
  return 0;
}

struct MotnTag * motn_root()
{
  return root;
}

void motn_apply(struct MotnTag * motn, 
                struct SktrTag * sktr,
                int PoseNo)
{
  int i;
  uchar * angles = motn->AngleList;
  int LimbCountMotn = motn->motn.LimbCount;
  int LimbCountSkel = sktr->skel->LimbCount;
  
  int offset = PoseNo;
  int delta = motn->motn.PoseCount;

  _CHECK(LimbCountSkel > LimbCountMotn,
         "limb count in skeleton is bigger than in motion %d %d", 
         LimbCountSkel, LimbCountMotn);

  _CHECK(PoseNo > motn->motn.PoseCount,
         "pose overflow %u in %s", PoseNo, motn->motn.name);

  #define XAA sktr->skel->LimbList[i].xaa
  #define YAA sktr->skel->LimbList[i].yaa
  #define ZAA sktr->skel->LimbList[i].zaa

  for (i = 0; i < LimbCountSkel; i++, offset += delta)
    XAA = (uchar) (angles[offset]);

  offset = delta * LimbCountMotn + PoseNo;
  for (i = 0; i < LimbCountSkel; i++, offset += delta)
    YAA = (uchar) (angles[offset]);

  offset = delta * LimbCountMotn * 2 + PoseNo;
  for (i = 0; i < LimbCountSkel; i++, offset += delta)
    ZAA = (uchar) (angles[offset]);

  #undef XAA
  #undef YAA
  #undef ZAA
}

void motn_start(struct MotnTag * motn,
                struct SktrTag * sktr,
                int PoseNo)
{
  memcpy(sktr->AQuaternionList,
         sktr->BQuaternionList, 
         QTRSIZE * sizeof(float) * sktr->skel->LimbCount);

  motn_apply(motn, sktr, PoseNo);
  skel_start(sktr);

  /*
   * sktr->BQuaternionList now contains pose delta rotations
   *
   */

  sktr->SlerpStep = 1;
}                

int motn_next(struct SktrTag * sktr)
{
  int     i;
  float   step;
  float * AQtr;
  float * BQtr;
  float * ItrpQtr;

  _CHECK(sktr->SlerpStep > sktr->SlerpStepCount,
         "slerp overflow %u %u",
         sktr->SlerpStep, sktr->SlerpStepCount);


  AQtr    = sktr->AQuaternionList;
  BQtr    = sktr->BQuaternionList;
  ItrpQtr = sktr->ItrpQuaternionList;

  if (sktr->SlerpStep == sktr->SlerpStepCount)
  {
    memcpy(ItrpQtr, BQtr, QTRSIZE * sizeof(float) * sktr->skel->LimbCount);
    sktr->SlerpStep++;
    return 1;
  }

  step = (float) sktr->SlerpStep / (float) sktr->SlerpStepCount;

  for (i = 0; i < sktr->skel->LimbCount; i++,
                                         ItrpQtr += 4, AQtr += 4, BQtr += 4)
  {
    qtr_slerp(ItrpQtr, AQtr, BQtr, step);
  }
  sktr->SlerpStep++;
  return 0;
}

