#include "config.h"

#include <stdio.h>
#include "qtr.h"

void qtr_qtrmul_Mikael_Kalms(float * r, float * a, float * b);
void qtr_qtrmul_Nick_Bobick(float * r, float * a, float * b);
void qtr_qtrmul_Henrik_Engstrom(float * r, float * a, float * b);

void qtr_qtrmul(float * r, float * a, float * b)
{
  qtr_qtrmul_Mikael_Kalms(r, a, b);
}

void qtr_qtrmul_Nick_Bobick(float * r, float * a, float * b)
{
  float A, B, C, D, E, F, G, H;

  A = (a[3] + a[0]) * (b[3] + b[0]);
  B = (a[2] - a[1]) * (b[1] - b[2]);
  C = (a[0] - a[3]) * (b[1] - b[2]);
  D = (a[1] + a[2]) * (b[0] - b[3]);
  E = (a[0] + a[2]) * (b[0] + b[1]);
  F = (a[0] - a[2]) * (b[0] - b[1]);
  G = (a[3] + a[1]) * (b[3] - b[2]);
  H = (a[3] - a[1]) * (b[3] + b[2]);
  r[3] =  B + (-E - F + G + H) / 2;
  r[0] =  A - ( E + F + G + H) / 2; 
  r[1] = -C + ( E - F + G - H) / 2;
  r[2] = -D + ( E - F - G + H) / 2;
}

void qtr_qtrmul_Mikael_Kalms(float * r, float * a, float * b)
{
  float v0[3], v1[3];
  float q1[4], q2[4];

  QTRCPY(q1, a);
  QTRCPY(q2, b);

  VECCPY(v0, q2);
  VECCPY(v1, q1);
  VECSCL(v0, q1[3]);
  VECSCL(v1, q2[3]);

  r[3] = q1[3] * q2[3] - VECDOT(q1, q2);
  VECADD(v0, v1);
  VECCPY(r, q1);
  VECCRS(r, q2);
  VECADD(r, v0);
}

void qtr_qtrmul_Henrik_Engstrom(float * r, float * a, float * b)
{
  float A, B, C, D;
  float x, y, z, w;
  float tmp_00, tmp_01, tmp_02, tmp_03, tmp_04, tmp_05, tmp_06, tmp_07, tmp_08, tmp_09;

  #define QTRVAR(q, X, Y, Z, W) X = (q)[0]; Y = (q)[1]; Z = (q)[2]; W = (q)[3]

  QTRVAR(a, A, B, C, D);
  QTRVAR(b, x, y, z, w);

  tmp_00 = (D - C) * (z - w);
  tmp_01 = (A + B) * (x + y);
  tmp_02 = (A - B) * (z + w);
  tmp_03 = (C + D) * (x - y);
  tmp_04 = (D - B) * (y - z);
  tmp_05 = (D + B) * (y + z);
  tmp_06 = (A + C) * (x - w);
  tmp_07 = (A - C) * (x + w);
  tmp_08 = tmp_05 + tmp_06 + tmp_07;
  tmp_09 = (float) 0.5 * (tmp_04 + tmp_08);

  x = tmp_00 + tmp_09 - tmp_05;
  y = tmp_01 + tmp_09 - tmp_08;
  z = tmp_02 + tmp_09 - tmp_07;
  w = tmp_03 + tmp_09 - tmp_06;

  QTRPAK(r, x, y, z, w);
}

/* 
 * q v q'
 */
void qtr_vecmul(float * r, float * q, float x, float y, float z)
{
  float qt[QTRSIZE], qi[QTRSIZE];

  QTRPAK(qt, x, y, z, 0.f);

  QTRCPY(qi, q);
  QTRINV(qi);

  qtr_qtrmul(qt,  q, qt);
  qtr_qtrmul(qt, qt, qi);

  VECCPY(r, qt);
}

void qtr_rotation(float * r, float theta, float x, float y, float z)
{
  float halftheta;
  float mag;

  halftheta = (float) (theta * DEG2RAD) / 2.f;
  VECPAK(r, x, y, z);

  mag = VECMAG(r);

  VECSCL(r, (float) (sin(halftheta) / mag));

  r[3] = (float) cos(halftheta);
}

void qtr_slerp(float * r, float * a, float * b, float t)
{
  float cosphi, phi;
  float k0, k1;
  float qt[QTRSIZE];
  int flip;

  cosphi = QTRDOT(a, b);

  /* if q1 is on opposite hemisphere from q0, use -q1 instead */
    
  flip = 0;
  if (cosphi < 0)
  {
    flip = 1;
    cosphi = -cosphi;
  }

  /* LERP when q0 is almost equal to q1 */
    
  if (1.f - cosphi < EPSILON)
  {
    k0 = 1.f - t;
    k1 = t;
  }
  else
  {
    phi = (float) acos(cosphi);
    k0 = (float) (sin((1.f - t) * phi) / sin(phi));
    k1 = (float) (sin(t * phi) / sin(phi));
  }

  if (flip)
    k1 = -k1;

  /* Just Do It. */
    
  QTRCPY(qt, a);
  QTRSCL(qt, k0);
  QTRCPY(r, b);
  QTRSCL(r, k1);
  QTRADD(r, qt);
}

void qtr_mtx(float * r, float * q)
{
  float s;
  float xs, ys, zs;
  float wx, wy, wz, xx, xy, xz, yy, yz, zz;

/*    
  float n;

  n = q_norm(q);
  s = (n > 0) ? (2 / n) : 0;
*/

/*  _CHECK(q_norm(q) > 0, "qtr_tomtx error", 0, 0);*/

  s = 2.f / QTRNRM(q);

  xs = q[0] * s;
  ys = q[1] * s;
  zs = q[2] * s;

  wx = q[3] * xs;
  wy = q[3] * ys;
  wz = q[3] * zs;
  xx = q[0] * xs;
  xy = q[0] * ys;
  xz = q[0] * zs;
  yy = q[1] * ys;
  yz = q[1] * zs;
  zz = q[2] * zs;

  r[0] = 1.f - yy - zz;
  r[1] = xy + wz;
  r[2] = xz - wy;

  r[3] = xy - wz;
  r[4] = 1.f - xx - zz;
  r[5] = yz + wx;

  r[6] = xz + wy;
  r[7] = yz - wx;
  r[8] = 1.f - xx - yy;
}

#ifdef _TEST

int main(int argc, char ** argv)
{
  return 0;
}

#endif
